from .base import Tag, on_render, TagConfig


# suppresses registering of hooks
class BlankUi:

    def __getattribute__(self, name):
        return lambda *_, **__: None

# REFACTOR

class RootTag(Tag):

    _config = TagConfig(name="root")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._ui = None

        self.entry_index = 0
        self.current_entry = None
        self.used_entry_chars = []
        
        self.rerender = False
        self.config = {}
        self.empty_entry = None

    def inject_ui(self, ui):
        self._ui = ui

        return self

    def double_render(self):
        self.rerender = True

    # special case, doesn't need decorator
    async def render(self):

        self.entry_count = 0
        self.used_entry_chars = []

        await self.render_children()
        
        # this could get dangerous
        if self.rerender:
            self.rerender = False
            await self.render()
        
        return self

    def next_entry(self):
        if self.entry_count > self.entry_index + 1:
            self.entry_index += 1

    def previous_entry(self):
        if self.entry_index > 0:
            self.entry_index -= 1

    def message_tags(self):

        messages = []
        stack = self.children

        while stack:
            child = stack.pop()

            if child.name == "message":
                messages.insert(0, child)

            else:
                stack += child.children

        return messages


import re



def addition(a, b):
    if str in (type(a), type(b)):
        a = str(a)
        b = str(b)

    return b + a


OPERATORS = {
    '+':  (addition, 1),
    '-':  (lambda a, b: b - a,  1),
    '*':  (lambda a, b: b * a,  2),
    '/':  (lambda a, b: b / a,  2),
    '==': (lambda a, b: b == a, 0),
    '!=': (lambda a, b: b != a, 0),
    '>':  (lambda a, b: b > a,  0),
    '>=': (lambda a, b: b >= a, 0),
    '<':  (lambda a, b: b < a,  0),
    '<=': (lambda a, b: b <= a, 0),
    '^':  (lambda a, b: b or a, 0),
    '&':  (lambda a, b: b and a, 0),
    '(':  (lambda *_: _, 0),
    ')':  (lambda *_: _, 0)
}

CONST = r"(?P<const>[0-9\.]+)"
STR   = r"\"(?P<str>.*?)\""
VAR   = r"(?P<var>[a-zA-Z_][a-zA-Z0-9_\. ]*[a-zA-Z0-9_]+)"
#                        this escape only applies to the first char in OP
OP    = r"(?P<op>" + '|'.join(f'\\{op}' for op in OPERATORS) + ")"

# this should happen on decorator
# currently defaults to None
def get_var(tag, var_name, default=None):
    name, *index = var_name.split('.')
    while True:
        # BUG: this causes the tag to look for state var all the
        # way up to root before even attempting to consider tag args
        # BUG: used to just check if var, this made it impossible to get 0
        if (var := tag.get_state(name)) is not None:
            # supports for chaining indexes??
            if index:
                index = int(index[0])
                # temp solution to ensure out of bounds errors are not thrown,
                # returns 0 for easier stats
                if len(var) > index:
                    return var[index]
                else:
                    return 0
            return var
        if (tag := tag.parent) is None:
            return default


def shunting_yard(tag, string):
    op_stack = []
    out_queue = []
    
    ex = f"{CONST}|{STR}|{OP}|{VAR}"

    try:

        for match in re.finditer(ex, string):
            const, string, op, var = match.groups()

            if const:
                out_queue.append(int(const))

            elif string:
                out_queue.append(string)

            elif var:
                v = get_var(tag, var)
                out_queue.append(v)

            elif op:
                if op == ')':
                    while (next_op := op_stack.pop())[0] != '(':
                        _, func, _ = next_op

                        out_queue.append(
                            func(out_queue.pop(), out_queue.pop())
                        )
                else:
                    func, precedence = OPERATORS[op]

                    if op_stack:
                        _, prev_func, prev_precedence = prev_op = op_stack.pop()

                        if op != '(' and prev_precedence >= precedence:
                            out_queue.append(
                                prev_func(out_queue.pop(), out_queue.pop())
                            )
                        else:
                            # append back the previous operator, if not used
                            op_stack.append(prev_op)

                    op_stack.append((op, func, precedence))

        for (op, func, _) in op_stack[::-1]:
            out_queue.append(
                func(out_queue.pop(), out_queue.pop())
            )

        return out_queue[0]
    
    except Exception as err:
        raise Exception(f"Failed to evaluate: \n{string}\n" + err)


if __name__ == '__main__':
    s = "((5 + 5) * 3 + 1 == 32) == 0"
    out = shunting_yard(None, s)
    print(out)


from __future__ import annotations

import discord
import re

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from src.hooks import HookTrigger


class Hook:

    def __init__(self, *_, checks=[lambda _: True]):
        self.checks = checks

    def matches(self, trigger: HookTrigger):
        if trigger.type != type(self):
            return False

        for check in self.checks:
            if not check(trigger):
                return False

        return True


class ReactionHook(Hook):

    def __init__(self, emoji: str, *_, **kwargs):
        super().__init__(**kwargs)

        self.emoji = emoji

    def matches(self, trigger: HookTrigger):
        if not super().matches(trigger):
            return

        # print(self.emoji, trigger.data, type(trigger.data))

        if self.emoji == trigger.data:
            return True


class MessageHook(Hook):

    def __init__(self, expression: str, *_, **kwargs):
        super().__init__(**kwargs)

        self.expression = expression

    def matches(self, trigger: HookTrigger):
        if not super().matches(trigger):
            return

        # print(self.expression, trigger.data, type(trigger.data))

        if match := re.fullmatch(self.expression, trigger.data, re.DOTALL):
            # print("MATCH")
            return True


class HookTrigger:

    def __init__(self, type: Hook, data: str, user: discord.User):
        self.type = type
        self.data = data
        self.user = user

        self.response = {}
        self.resolved = False

    def delete_input(self):
        return self.response.get('delete_input', True)

    def resolve(self, response=None):
        self.resolved = True
        if response:
            self.response = response

    def is_resolved(self) -> bool:
        return self.resolved



import re

from .base import TAGS 

from .root import RootTag
from .base import *
from .control import *


def new_tag(name, *args, **kwargs):
    if cls := TAGS.get(name):
        return cls(*args, **kwargs)

    raise Exception(f"Unknown tag '{name}'")


# removes leading and trailing whitespaces and tabs
def normalize_form(form) -> str:

    lines = []

    for line in form.split('\n'):

        match = re.fullmatch(r"[ \t]*(.*?)[ \t]*", line)

        lines.append(match.group(1))

    return '\n'.join(lines)


def parse(form: str) -> RootTag:
    # stacks opening tags and pops them using closing tags 
    stack = []

    # we are changing form during iteration so we need to
    # keep track of how many characters have been offset
    offset = 0

    form = normalize_form(form)
    
    # matches either opening or closing tag
    # cannot handle newlines here bc tags would "compete" for the newline
    for match in re.finditer(r"(\n?)<(/?)([a-zA-Z0-9]+)(.*?)(/?)>", form):

        # singleton tags should be explicitly defined in TagConfig
        # this would simplify the algorithm a lot
        br, closing, name, kwargs, contentless = match.groups()

        # avoids matching discord formatting elements such as <t:timestamp>
        if kwargs:
            if kwargs[0] == ':':
                continue
        
        # adds opening tags to stack along with its match start and end
        if not closing:
            tag = new_tag(name, kwargs, stack[-1][0] if stack else None)

            singleton = tag.singleton
            contentless = contentless or (not tag._config.content)
            # singletons will also be appened to stack, but are popped immediatelly
            stack.append((tag, match.start() - offset, match.end() - offset))

        # handles the tag closing
        if closing or singleton:
            top, start, end = stack.pop()

            # ensures the form is valid by checking if the closing tag matches the
            # top of the stack
            if top.name != name:
                raise SyntaxError(
                    f"Incompatible opening tag {top.name} with closing tag /{name}"
                )

            # the content of the tag is the string between match end of the opening tag
            # and the match start of the closing tag, while accounting for offset
            content = form[end:match.start() - offset]

            # strips leading and trailing newlines if present
            # to sanitize tag content
            # if not contentless and content:
            #     if content[0] == '\n':
            #         content = content[1:]

            #     if content[-1] == '\n':
            #         content = content[:-1]

            top.set_content(content)

            # the original form is stripped of the tag and it's content
            # and replaced with {} which will later be formatted by the 
            # tag itself at render time to produce dynamic content
            form = form[:start] + \
                ("" if contentless else "{}") + \
                form[match.end()-offset:]

            # offset keeps track of how many characters were removed from the form
            # by taking the index of the end of the closing tag and decrementing the
            # start index of the opening tag, then adding 2 if the tag isn't contentless
            # since we add 2 characters, 
            # but since the start and end indexes aren't dynamic
            # and do not account for the form changes we need to decrement offset here
            # to account for form changes manually 
            offset += match.end() - offset - start - (0 if contentless else 2)
            
            # if the stack isn't empty the current tag
            # will be appended to the top of the stack as child
            if stack:
                stack[-1][0].children.append(top)
            # if the stack is empty we are currently handling the closing part
            # of the root element and we can return it and finish parsing
            else:
                return top



import yaml
import re

from yaml import Loader

from ..hooks import HookTrigger, MessageHook, ReactionHook
from ..utils import find_item_in_array

from .base import Tag, on_render, TagConfig, Optional, Union, EmptyEntry, InvalidEntry


class ConfigTag(Tag):

    _config = TagConfig(name="config")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = {}

    @on_render()
    async def render(self, args):
        self.config = yaml.load(self.content, Loader=Loader)
        self.root.config = self.config


class ForTag(Tag):

    _config = TagConfig(
        name="for",
        args={
            "in":     list,
            "var":    Optional(str, "item"),
            "start":  Optional(int, None),
            "end":    Optional(int, None),
            "inline": Optional(bool, False)
        }
    )

    @on_render()
    async def render(self, args):
        out = []
        items = args['in'] or []
        
        for index, item in enumerate(items[args.start:args.end]):
            self.update_state({
                args.var: item,
                'index': index + 1
            })
            if to_add := await self.render_content():
                out.append(to_add)

        return (' ' if args.inline else '').join(out)


class IfTag(Tag):

    _config = TagConfig(
        name="if",
        args={
            "exp": bool
        }
    )

    @on_render()
    async def render(self, args):
        if args.exp and (content := await self.render_content()):
            return content
            
        return ""


# might be best to create an action for this
class CheckTag(Tag):

    _config = TagConfig(name="check", singleton=True)

    @on_render()
    async def render(self, args):
        async def event(*_):
            if entry := self.root.empty_entry:
                raise EmptyEntry(entry.get_arg('name'))

        self.ui.queue_event(event)


class EntryTag(Tag):

    _config = TagConfig(
        name="entry",
        args={
            "name":    str,
            "var":     Optional(str),
            "require": Optional(bool, False),
            "value":   Optional(str),
            "type":    Optional(str),
            "integer": Optional(bool, False),
            "options": Optional(str),
            "match":   Optional(str),
            "show":    Optional(bool, True),
            "repeat":  Optional(bool, True),
            "move":    Optional(bool, True)
        }
    )

    MATCHES = {
        'tag': r"\<@!?([0-9]+?)\>",
        'role': r"<@&([0-9]+?)>",
        'channel': r"<#([0-9]+?)>",
    }

    def parse_data(cls, new_data, old_data, args):

        # hacky shit here, need better parsing
        if exp := cls.MATCHES.get(args.match):
            if not (data := re.findall(exp, new_data)):
                raise InvalidEntry(f"Input of type \"**{args.match}**\" required.")
                 
        elif args.integer:
            try:
                data = [int(new_data)]
            except (ValueError, TypeError):
                raise InvalidEntry(f"Integer input required.")
                
        else:
            data = [new_data]
            
        if any(opts := re.split(r", ?", args.options or "")):
            opt_checked_data = [] 

            for index, item in enumerate(data):

                if checked_item := find_item_in_array(item, opts, threshold=0.05):

                    opt_checked_data.append(checked_item)

            if opt_checked_data:
                data = opt_checked_data
            else:
                raise InvalidEntry(f"Invalid option \"**{new_data}**\"")

        if args.type == 'list':

            if not args.repeat:

                non_repeat_data = old_data or []

                for item in data:
                    if item not in non_repeat_data:
                        non_repeat_data.append(item)

                data = non_repeat_data

            else:
                data = (old_data or []) + data
        
        else:
            data = data[0] if data else ''

        return data

    # method for setting entry data; triggered by message hook ([^\-].+)
    async def set_data(self, trigger, args):
        
        old_data = self.root.get_state(args.name) or []
        parsed_data = self.parse_data(trigger.data, old_data, args)

        self.root.update_state({args.name: parsed_data})
        self.update_state({"length": len(parsed_data)})

        if args.move and parsed_data and (old_data != parsed_data):
            self.root.next_entry()

        trigger.resolve({"delete_input": True})

    # method for deleting entry data; triggered by message hook (-d)
    async def delete_data(self, trigger, args):
        self.root.update_state({args.name: ''})
        self.update_state({"length": 0})

        trigger.resolve({"delete_input": True})

    # method for deleting the last item in an array entry field;
    # triggered by message hook (-)
    #
    # ASSUMES THE ENTRY DATA IS A LIST!!
    async def delete_last(self, trigger, args):
        data = self.root.get_state(args.name)

        self.root.update_state({args.name: data[:-1]})

        self.update_state({"length": len(data) - 1})

    # method for quick navigation between entries; 
    # triggered by message hook using a letter from the entry name
    async def navigate_to(self, trigger, index):
        self.root.entry_index = index

        trigger.resolve()
        # why return true?
        return True

    def message_hook(self, exp, func, *args):
        self.ui.register_hook(
            MessageHook(exp, checks=[
                lambda t: t.user.id == self.ui.renderer.author.id
            ]),
            func=lambda trigger: func(trigger, *args)
        )

    def register_nav_char(self, name):

        name_chars = list(name)
        index   = 0
        discrim = 0

        while name_chars and (ch := name_chars.pop(0)):

            if (l_ch := ch.lower()) not in self.root.used_entry_chars:

                self.update_state({
                    'dname': f"{name[:index]}__{ch}__{name[index+1:]}"
                })
                self.root.used_entry_chars.append(l_ch)
                break

            if not name_chars:
                name_chars.append(str(discrim := discrim + 1))

            index += 1

        # register quick navigation hook
        self.message_hook(
            f"({ch.lower()}|{ch.upper()})", self.navigate_to, self.root.entry_count
        )

    def register_set_data_event(self, args):
        async def wrapper():
            self.message_hook(r"([^\-].*)", self.set_data, args)

        return wrapper

    # checks whether the given entry is required,
    # if so it's set as an emtpy_entry on the root
    # NOT TESTED
    def check_required(self, data, args):
        # to prevent a situation where an actual entry is skipped because the empty_entry
        # slot is being hogged by a non-empty entry, every entry attempts to set itself as THE empty_entry;
        # this results in empty entry errors resolving up backwards
        if args.required and not data:
            self.root.empty_entry = self

        elif data and self.root.empty_entry == self:
            self.root.empty_entry = None

    @on_render()
    async def render(self, args):

        entry_data = self.root.get_state(args.name)

        self.check_required(entry_data, args)

        self.update_state({
            # this lets the placeholder value to be used as possible argument
            # in other tags, leading to undesired behaviour
            args.var or 'value': entry_data or args.value,
            'name':   args.name,
            'dname':  args.name,
            'active': 0
        })

        self.register_nav_char(args.name)

        # check whether this entry is active
        if self.root.entry_index == self.root.entry_count:

            self.update_state({
                'dname':  f"[ {self.get_state('dname')} ]",
                'active': 1
            })
            # register quick nav hook
            self.message_hook(r"([^\-].+)", self.set_data, args)

            # register hook to delete entry data
            self.message_hook(r"(-d)", self.delete_data, args)

        self.root.entry_count += 1

        # only shows the entry if it's active or 'show' is true
        if args.show or self.get_state('active'):
            return await self.render_content()


class SelectEntry(Tag):

    _config = TagConfig(
        name="select",
        args={
            "index": int
        },
        singleton=True
    )

    def select_by_index(self, index):
        async def wrapper():
            self.root.entry_index = index

        return wrapper


    @on_render()
    async def render(self, args):
        self.ui.queue_event(self.select_by_index(index))


# sets/changes a root state variable
class SetTag(Tag):

    _config = TagConfig(
        name="set",
        args={
            "var": str,
            "value": Union(str, int, list)
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.root.update_state(
            {args.var: args.value}
        )
        # page flipping needs this to work but 
        # when outside conditional it causes infinte loop
        # self.root.double_render()


class GetTag(Tag):

    _config = TagConfig(
        name="get",
        args={
            "var":     str,
            "source":  list,
            "key":     str,
            "default": Optional(int, 0)
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        for item in args.source:
            if item[0] == args.key:
                var = item[1:]
                break
        else:
            var = args.default

        self.root.update_state({args.var: var})


class DecorTag(Tag):

    _config = TagConfig(
        name="decor",
        args={
            "exp": Optional(bool, False),
            "op":  str,
            "ed":  str
        }
    )

    @on_render()
    async def render(self, args):
        out = await self.render_content()

        if args.exp:
            out = args.op + out + (args.cl or args.op[::-1])
        
        return out



import discord

from discord import Embed
from datetime import datetime

from .base import Tag, on_render, TagConfig, Optional

from ..hooks import ReactionHook, HookTrigger


EMOJIS = {
    'up': "⬆",
    'down': "⬇",
    'checkmark': "✅",
    'cross': "<:red_cross_5:744249221011472444>",
    'file': "📁"
}


class Message(Tag):

    _config = TagConfig(name="message")

    def add_reaction(self, emoji, event=None):
        self.reactions.append(emoji)

    @on_render()
    async def render(self, args):
        self.embed = None
        self.reactions = []
        self.message = await self.render_content()


class EmbedTag(Tag):

    _config = TagConfig(
        name="embed",
        content=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # this might eventually cause issues,
        # might be best to create separate API for managing embeds
        self.embed = discord.Embed()

    @on_render()
    async def render(self, args):
        self.embed.clear_fields()
        self.reactions = []
        await self.render_children()
        
        self.first_ancestor('message').embed = self.embed

        return ""


class DescriptionTag(Tag):

    _config = TagConfig(name="description", content=False)

    @on_render()
    async def render(self, args):
        descption = await self.render_content()
        self.first_ancestor("embed").embed.description = descption 


class TitleTag(Tag):

    _config = TagConfig(name="title", content=False)

    @on_render()
    async def render(self, args):
        title = await self.render_content()  
        self.first_ancestor("embed").embed.title = title
        

class ColorTag(Tag):

    _config = TagConfig(
        name="color",
        args={
            "value": str
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.first_ancestor("embed").embed.color = int(args.value, 16)


class FieldTag(Tag):

    _config = TagConfig(
        name="field",
        args={
            "name":   str,
            "inline": Optional(bool, False)
        }
    )

    @on_render()
    async def render(self, args):
        self.first_ancestor("embed").embed.add_field(
            name=args.name, 
            inline=args.inline, 
            value=await self.render_content()
        )


class ImageTag(Tag):

    _config = TagConfig(
        name="image",
        args={
            "url":   str,
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.first_ancestor("embed").embed.set_image(url=args.url)


class ThumbnailTag(Tag):

    _config = TagConfig(
        name="thumbnail",
        args={
            "url":   str,
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.first_ancestor("embed").embed.set_thumbnail(url=args.url)


class FooterTag(Tag):

    _config = TagConfig(
        name="footer",
        args={
            "url": Optional(str, ""),
        }
    )

    @on_render()
    async def render(self, args):
        self.first_ancestor("embed").embed.set_footer(
            icon_url=args.url, 
            text=await self.render_content()
        )


class TimestampTag(Tag):

    _config = TagConfig(
        name="timestamp",
        args={
            "value": int
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        if args.value:
            ts = datetime.fromtimestamp(args.value)
            self.first_ancestor("embed").embed.timestamp = ts


class CodeTag(Tag):

    _config = TagConfig(
        name="code", 
        args={
            "lang": Optional(str, "")
        }
    )

    @on_render()
    async def render(self, args):
        content = await self.render_content()

        return f"```{args.lang}\n{content}```"


class ReactionTag(Tag):

    _config = TagConfig(
        name="r",
        args={
            "emoji": str
        }
    )
    
    async def on_reaction(self, trigger: HookTrigger):
        self.update_state({
            'value': 1,
            'user': trigger.user.id
        })
        
        self.rendered_content = await self.render_content()

        trigger.resolve()

    @on_render()
    async def render(self, args):

        emoji = EMOJIS[args.emoji]
        
        self.first_ancestor('message').add_reaction(emoji)

        self.ui.register_hook(
            ReactionHook(
                emoji,
                checks=[lambda t: t.user.id == self.ui.renderer.author.id]
            ), 
            func=self.on_reaction
        )

        if self.get_state('value') == 1:

            self.update_state(value=0)

            return self.rendered_content


class ClassifiedTag(Tag):

    _config = TagConfig(
        name="c",
        args={
            "role": str
        }
    )

    @on_render()
    async def render(self, args):
        context = self.ui.renderer.context
        member = context.author

        if int(args.role) in [r.id for r in member.roles]:
            return await self.render_content()
        else:
            return '\n[ REDACTED ]'


# refactoring still needed here 
class UserContextTag(Tag):

    _config = TagConfig(
        name="user",
        args={
            "id": str
        }
    )

    def get_member_info(cls, member):
        return [
            member.name + '#' + member.discriminator,
            member.nick or member.name,
            [r.id for r in member.roles],
            str(member.avatar_url)
        ]

    @on_render()
    async def render(self, args):
        context = self.ui.renderer.context

        if user_id := args.id:

            try:
                user_id = int(user_id)
            except (TypeError, ValueError):
                return ''

            context = self.ui.renderer.context
            member_info = []

            if fetchUsers := self.first_ancestor('fetchUsers'):
                
                # member is not guaranteed to exist here
                if member := fetchUsers.get_state(user_id):
                    # saves the member info in root for future use,
                    # in case the member leaves the server
                    self.root.update_state({
                        user_id: self.get_member_info(member)
                    })

                    member_info = self.get_member_info(member)

                else:
                    return ''

            elif member := await context.guild.fetch_member(user_id):
                member_info = self.get_member_info(member)

            elif member := self.root.get_state(user_id):
                member_info = member

            if member_info:
                self.update_state({
                    'name':       member_info[0],
                    'nick':       member_info[1] or member_info[0],
                    'roles':      member_info[2],
                    'avatar_url': member_info[3],
                    'user':       member if type(member) is discord.Member else None
                })

            else:
                self.update_state({
                    'name':       None,
                    'nick':       None,
                    'roles':      [],
                    'avatar_url': None,
                    'user':       None
                })

            return await self.render_content()
            
        return ''


class FetchMembers(Tag):

    _config = TagConfig(name="fetchUsers")

    @on_render()
    async def render(self, args):
        context = self.ui.renderer.context

        members = []

        async for member in context.guild.fetch_members():
            members.append(member)

            self.update_state({member.id: member})

        return await self.render_content()


class ChannelContext(Tag):

    _config = TagConfig(
        name="channelContext",
        args={
            "id": int
        }
    )

    @on_render()
    async def render(self, args):

        context = self.ui.renderer.context

        channel = context.guild.get_channel(channel_id=args.id)

        members = []
        roles   = []

        for target, overwrite in channel.overwrites.items():

            if type(target) is discord.Member:
                members.append(target.id)

            elif type(target) is discord.Role:
                roles.append(target.id)

        self.update_state({
            "channel": channel,
            "hashtag": f"<#{args.id}>" if id else None,
            "ch_members": members,
            "ch_roles":   roles
        })

        return await self.render_content()


class SetChannelPerms(Tag):

    _config = TagConfig(
        name="chSetPerms",
        args={
            "channel": discord.TextChannel,
            "overwrite": int,
            "member": Optional(discord.Member),
            "role": Optional(discord.Role)
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):

        if args.member:
            target = args.member

        elif args.role:
            target = args.role

        else:
            raise Exception("Either member or role must be specified.")

        await args.channel.set_permissions(
            target, 
            overwrite=discord.PermissionOverwrite(
                # gotta do this hacky shit because discord.py
                **{k:v for k,v in discord.Permissions(args.overwrite)}
            )
        )
        


import asyncio
import discord

from discord import Embed

from .listener import DiscordListener

from .tags.base import Tag

from .hooks import HookTrigger
from .hooks import ReactionHook, MessageHook


class TemplateDisplay:

    def __init__(self, context):
        self.context = context

        self.messages: [discord.Message] = []
        # discord.py returns empty list when attempting to retrieve message reactions,
        # so we need to manually keep track of all the reactions;
        # the issue with this approach is that if someone were to manually
        # delete the bot reaction it wouldn't be automatically repaired
        self.reactions: {int: [str]} = {}

    # returns all message tags from root;
    # should be handled in root
    def message_tags(self, root) -> [Tag]:
        return [t for t in root.children if t.name == 'message']
        # return root.message_tags()

    def get_message_ids(self) -> [int]:
        return [message.id for message in self.messages]

    async def update_reactions(self, message, reactions):
        # check if reactions need updating
        if reactions == (curr_reactions := self.reactions.get(message.id, [])):
            return

        if curr_reactions:
            await message.clear_reactions()
        # manually keeping track of the reactions we are adding
        self.reactions[message.id] = reactions

        for emoji in reactions:
            await message.add_reaction(emoji)

    # creates and publishes a new message object while adding
    # it to the messages list
    async def new_message(self, tag) -> discord.Message:
        message = await self.context.channel.send(content=tag.message, embed=tag.embed)

        self.messages.append(message)

        return message

    # retrieves message object from messages list
    def get_message(self, index) -> discord.Message:
        message = None

        if index <= len(self.messages) - 1:
            message = self.messages[index]

        return message

    # checks if there are any leftover message objects 
    # that do not match to any message tags and deletes them
    async def message_leftover_cleanup(self, root):
        del_message_tasks = []

        # deletes all the messages with index of len(tags) and higher
        for message in self.messages[len(self.message_tags(root)):]:
            del_message_tasks.append(asyncio.create_task(message.delete()))

            self.messages.pop(index + 1)

        if del_message_tasks:
            await asyncio.gather(*del_message_tasks)

    # first post all embeds then add reactions
    async def construct(self, root):
        reactions = []

        for tag in self.message_tags(root):

            message = await self.new_message(tag)

            reactions.append(asyncio.create_task(
                self.update_reactions(message, tag.reactions)
            ))

        await asyncio.gather(*reactions)

    # updates the message objects with new content and embed data
    # and also updates the reactions if needed
    async def update(self, root):
        reactions = []

        for index, tag in enumerate(self.message_tags(root)):

            if message := self.get_message(index):
                await message.edit(content=tag.message, embed=tag.embed)
            else:
                message = await self.new_message(tag)
        
            reactions.append(asyncio.create_task(
                self.update_reactions(message, tag.reactions)
            ))

        await self.message_leftover_cleanup(root)
        await asyncio.gather(*reactions)

    # deletes all messages
    async def discard(self):
        tasks = []

        for message in self.messages:
            tasks.append(asyncio.create_task(message.delete()))

        await asyncio.gather(*tasks)

    # deletes all reactions; redundant?
    async def deactivate(self):
        for message in self.messages:
            await self.update_reactions(message, [])


class DiscordRenderer:

    def __init__(self, context):
        self.context = context

        # might rename this to event_receiver?
        self.trigger_matcher = None
        self.trigger_resolver = None
        
        self.display = TemplateDisplay(context)
        self.listener = DiscordListener()

        self.swap = False
        self.queued_error = None
        self.error_message = None

    @property
    def server(self) -> discord.Guild:
        return self.context.guild

    @property
    def channel(self) -> discord.TextChannel:
        return self.context.channel

    @property
    def author(self) -> discord.Member:
        return self.context.author

    @property
    def messages(self) -> [int]:
        return self.display.get_message_ids()

    # meh, would be best to find a way around this
    def inject_trigger_matcher(self, matcher):
        self.trigger_matcher = matcher

    def inject_trigger_resolver(self, resolver):
        self.trigger_resolver = resolver

    def match_trigger(self, trigger):
        return self.trigger_matcher(trigger)

    # pushes the trigger into interface's receive_event loop
    # and waits for a response; 
    async def resolve_trigger(self, trigger) -> HookTrigger:
        # if event_dispatcher returns it indicates 
        # that the event loop has ended 
        if await self.trigger_resolver():
            # returning None signals to listener that the template
            # was disabled and clean-up is not to be attempted
            #
            # this will however prevent messages from getting cleaned up,
            # if we ever try to terminate the template with message input
            return

        # trigger should change to ResolvedTrigger to indicate
        # that it has been resolved
        if trigger.is_resolved():
            return trigger

    # eh
    def send_error(self, error):
        self.queued_error = self.context.channel.send(
            embed=discord.Embed(description=error, color=int('cc2040', 16))
        )

    # gets the dm_channel with the author,
    # creates a new channel if it doesn't exist
    async def get_dm_channel(self) -> discord.TextChannel:
        if not self.context.author.dm_channel:
            await self.context.author.create_dm()

        return self.context.author.dm_channel

    async def construct(self, root, register_listener=True):
        # this is stupid
        if root.config.get('channel') == 'dm' and not self.swap:
            # this is even stupider
            self.context.channel = await self.get_dm_channel()

        await self.display.construct(root)

        if register_listener:
            self.listener.add_renderer(self)

    # needs to be capable of swiching channels 
    async def update(self, root):

        # meh, refactoring still needed
        if self.error_message:
            await self.error_message.delete()
            self.error_message = None

        if self.queued_error:
            self.error_message = await self.queued_error
            self.queued_error = None

        if root.config.get('refresh') == 'resend':
            await self.display.construct(root)
        else:
            await self.display.update(root)

    def swap_channel(self, channel_id):
        self.swap = True
        channel = self.context.guild.get_channel(int(channel_id))

        # BAD, NO
        self.context.channel = channel

        self.display = TemplateDisplay(self.context)

    async def deactivate(self):
        self.listener.remove_renderer(self)
        await self.display.deactivate()

        self.display = None

    async def discard(self):
        self.listener.remove_renderer(self)
        await self.display.discard()

        self.display = None



import re

from abc import abstractmethod

from ..safeval import shunting_yard

# REFACTOR

TAGS = {}

class EmptyEntry(Exception):
    pass


class InvalidEntry(Exception):
    pass


class Optional:
    def __init__(self, _type, default=None):
        self._type = _type
        self.default = default


class Union:
    def __init__(self, *types):
        self.types = types


class TagArgs:
    def __init__(self):
        self.kwargs = {}

    def add(self, key, value):
        self.kwargs[key] = value

    def __getattr__(self, key):
        return self.kwargs.get(key)

    def __getitem__(self, key):
        return self.kwargs.get(key)

    def __iter__(self):
        return iter(self.kwargs)

    def keys(self):
        return self.kwargs.keys()


class TagConfig:

    def __init__(self, 
        name: str, 
        args: dict = None, 
        singleton: bool = False,
        content: bool = True
    ):
        self.name = name
        self.args = args or {} 
        self.singleton = singleton
        self.content = content


def try_converting_var(val, _type):
    try:
        val = _type(val)
    except TypeError:
        raise Exception(f"Incorrect data type, got: {type(val)}, expected: {_type}")

    return val


def get_value(self, arg_name, arg_type):
    if arg_name in self._static_kwargs:
        value = self._static_kwargs[arg_name]
    
    elif arg_name in self._dynamic_kwargs:
        ref = self._dynamic_kwargs[arg_name]
        value = shunting_yard(self, ref)

    else:
        if isinstance(arg_type, Optional):
            value = arg_type.default
        else:
            raise Exception(f"Missing required argument: {arg_name}")

    return value


def validate_value(value, arg_type):
    if type(arg_type) == Optional:
        if type(value) != arg_type._type and value:
            value = try_converting_var(value, arg_type._type)

    elif type(arg_type) == Union:
        if type(value) not in arg_type.types:
            raise Exception("Incorrect data type")

    else:
        if type(value) != arg_type:
            value = try_converting_var(value, arg_type)

    return value

# handles argument calculation and type annotations 
#
# too much is happening here, BAD
#
# args should be parsed and resolved on init,
# and passed as args object on render for convenience
def on_render():
    def wrapper(func):
        async def inner(self, *args, **kwargs):
            args_to_pass = TagArgs()

            for arg_name, arg_type in self._config.args.items():

                val = get_value(self, arg_name, arg_type)

                # val can be 0, so this must be None
                if val is None:
                    args_to_pass.add(arg_name, None)
                    continue

                val = validate_value(val, arg_type)
                
                args_to_pass.add(arg_name, val)

            return await func(self, args_to_pass)
        return inner
    return wrapper


class MetaTag(type):

    def __new__(cls, class_name, bases, attrs):
        new_class = type.__new__(cls, class_name, bases, attrs)

        config = attrs['_config']

        TAGS[config.name] = new_class

        return new_class
        

# need to return 'get_arg' back
class Tag(metaclass=MetaTag):

    _config = TagConfig(
        name="base"
    )

    def __init__(self, kwargs: str, parent = None):
        self.parent = parent
        self.children = []

        self._content = None
        self._state_vars = {}

        # holds {key: value} pairs
        self._static_kwargs = {
            k: v for k,v in re.findall(
                r"([a-zA-Z0-9]*?)=[\"](.*?)[\"]", kwargs
            )
        }
        # holds {key: reference} pairs
        # only available at render time
        self._dynamic_kwargs = {
            k: r for k,r in re.findall(
                r"([a-zA-Z0-9]*?)={(.*?)}", kwargs
            )
        }

    @classmethod
    @property
    def name(cls):
        return cls._config.name

    @classmethod
    @property
    def singleton(cls):
        return cls._config.singleton

    @property
    def root(self):
        tag = self
        while (parent := tag.parent):
            tag = parent
        return tag

    @property
    def ui(self):
        return self.root._ui

    def first_ancestor(self, name):
        tag = self.parent
        while tag and tag.name != name:
            tag = tag.parent
        return tag

    def get_state(self, key, default=None):
        return self._state_vars.get(key, default)

    def update_state(self, _dict: dict=None, **kwargs):
        if _dict:
            kwargs = {**_dict, **kwargs}
        self._state_vars = {**self._state_vars, **kwargs}

    def parse_text(self, text):

        # this doesn't need to be here
        def repl(match):
            if (result := shunting_yard(self, match[1])) is None:
                return ""

            return str(result)

        return re.sub(r"{([^\{\}]+?)}", repl, text)

    async def render_content(self):
        return self.content.format(*(await self.render_children()))

    # might not be necessary
    @property
    def content(self):
        return self.parse_text(self._content)

    def set_content(self, content):
        if not content:
            self._content = ''
        start = 0
        end = None
        self._content = content[start:end]

    async def render_children(self) -> list:
        children_contents = []

        for child in self.children:

            content = await child.render()

            if not child.singleton:
                children_contents.append(
                    # without this we would have 'None' whenever
                    # a child does not return any content
                    content or ''
                )

        return children_contents

    @abstractmethod
    async def render(self, *_):
        ...



from types import CoroutineType

from .tags import parse, RootTag, InvalidEntry, EmptyEntry
from .hooks import HookTrigger, Hook


# needs better API 
# - passing external variables into it (state managment)


class GeneralInterface:

    def __init__(self, renderer):
        self.renderer = renderer
        self.renderer.inject_trigger_resolver(self.execute_hook_func)
        self.renderer.inject_trigger_matcher(self.match_trigger_to_hook)

        self.root: RootTag = None

        self.input_hooks = []
        self.event_queue = []

        self.__suppress_update = False

    # generates a root by parsing the template form
    # and renders the template 
    async def construct(self, form: str, vars: dict={}):
        self.root = parse(form).inject_ui(self)

        if vars:
            self.root.update_state(**vars)

        await self.root.render()
        await self.renderer.construct(
            # only register listener if mode != "view"
            self.root, register_listener=vars.get("mode") != "view"
        )

    # not really sure where this is even used
    async def update(self):
        await self.root.render()
        await self.renderer.update(self.root)

    # api for registering input hooks from tags
    def register_hook(self, hook: Hook, func: CoroutineType):
        self.input_hooks.append((hook, func))

    # api for queueing events from tags
    def queue_event(self, event: CoroutineType):
        self.event_queue.append(event)

    # need to separate this from hook execution
    async def execute_events(self):
        while self.event_queue:
            event = self.event_queue.pop(0)
            try:
                # events which return True terminate event loop 
                # and suppress HookTrigger cleanup
                if await event(): return True
            # prevents report from being saved if required fields are empty;
            # not sure if here is the right place to handle this
            except EmptyEntry as entry_name:
                self.renderer.send_error(f"Field **'{entry_name}'** is required.")

    # prevents the template from visually updating during 
    # the update loop following the call of this method 
    #
    # rename to suppress_update pls
    def suppress_refresh(self):
        self.__suppress_update = True

    # checks if any of the input hooks matches the passed hook trigger
    # and schedules the hook function for execution if match
    def match_trigger_to_hook(self, trigger) -> bool:
        # goes through input_hooks backwards, this is to ensure 
        # entry hooks are handled correctly when changing entry mid render
        for hook, func in self.input_hooks[::-1]:

            if hook is None or hook.matches(trigger):
                # this could get dangerous, as func will begin execution
                # and might finish before it's awaited
                self.hook_func = func(trigger)
                return True

        return False

    # The way this currently works is input hooks and event queue get populated
    # during construction, however the first set of events is never executed
    # as the second render overrides them, leading to undesired behaviour
    async def execute_hook_func(self):

        # input_hooks and event_queue are purged and refilled here
        self.input_hooks = []

        if self.hook_func:
            try:
                await self.hook_func

            except InvalidEntry as error_message:
                # idk if this should be handled here
                self.renderer.send_error(error_message)

            self.hook_func = None
        
        # repopulates both input_hooks and event_queue
        await self.root.render()

        # events and empty input hooks can trigger deactivation of the template,
        # possibly leaving behind user input that only gets cleaned up after this method finishes,
        # this might be an issue when it comes to message commands later

        # some input hooks can be directly responsible for the executed events,
        # e.g. reaction triggering save/discard
        if await self.execute_events():
            # this is needed as events use it to signal
            # to the listener when an event terminates the template
            return True

        # deactivate interface if there are no input_hooks after rerender
        if not self.input_hooks:
            await self.renderer.deactivate()
            # suppresses HookTrigger cleanup
            return True

        if self.__suppress_update:
            self.__suppress_update = False
            return

        await self.renderer.update(self.root)
        
    
    # API FOR ROOT

    # deactivates the template without deleting it,
    # usually used after saving when done creating report
    def clean_up(self):
        async def __clean_up():
            await self.renderer.update(self.root)
            await self.renderer.deactivate()

            self.suppress_refresh()
            return True

        self.queue_event(__clean_up)

    # deletes and deactivates the template without saving
    def discard(self):
        async def __discard():
            await self.renderer.discard()
            self.suppress_refresh()

            return True

        self.queue_event(__discard)

    # replaces the current template with a new one
    # using the same renderer to display it
    def swap(self, form):
        async def __swap():
            self.root = parse(form).inject_ui(self)
            await self.root.render()

        self.queue_event(__swap)

    # moves the template to a new channel
    def forward(self, channel_id, mode=None):
        async def __forward():
            self.renderer.swap_channel(channel_id)
            await self.renderer.construct(self.root)

            # needs to be handled more generally elsewhere
            if mode == 'view':
                await self.renderer.deactivate()

            self.suppress_refresh()

        self.queue_event(__forward)


from .base import Tag, on_render, TagConfig, Optional


    
class MoveTag(Tag):

    _config = TagConfig(
        name='move',
        args={
            'dir': str,
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        if args.dir == 'up':
            self.root.previous_entry()

        elif args.dir == 'down':
            self.root.next_entry()

        # there might be a cleaner way to handle this
        # self.root.double_render()


class DiscardTag(Tag):

    _config = TagConfig(
        name='discard',
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.ui.discard()


class CleanupTag(Tag):

    _config = TagConfig(
        name='cleanup',
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.ui.clean_up()


class ForwardTag(Tag):

    _config = TagConfig(
        name="forward",
        args={
            "channel": int,
            "mode":    Optional(str, "view")
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        self.ui.forward(args.channel, args.mode)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="xmlcord",
    version="1.0.21",
    author="Fliqqr",
    author_email="author@example.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Fliqqr/xmlcord",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'aiohttp',
        'discord'
    ],
    python_requires='>=3.10',
)


from .base import Tag, on_render, TagConfig, Optional


def md(*mds):
    return f"\u001B[{';'.join(str(md) for md in mds)}m"

END = "\u001B[0m"

colors = { 
    "gray":   30,
    "red":    31,
    "green":  32,
    "yellow": 33,
    "blue":   34,
    "pink":   35,
    "cyan":   36,
    "white":  37,
}

backgrounds = {
    "darkblue": 40,
    "organe":   41,
    "gray3":    42,
    "gray2":    43,
    "gray1":    44,
    "indigo":   45,
    "gray0":    46,
    "white":    47,
}


class Bold(Tag):

    _config = TagConfig(name="b")

    @on_render()
    async def render(self, args):
        return md(1) + await self.render_content() + END


class Underline(Tag):

    _config = TagConfig(name="u")

    @on_render()
    async def render(self, args):
        return md(4) + await self.render_content() + END


class Tab(Tag):

    _config = TagConfig(
        name="t",
        args={
            "i": Optional(int, 1)
        },
        singleton=True
    )

    @on_render()
    async def render(self, args):
        return "    " * args.i


class StyleTag(Tag):

    _config = TagConfig(
        name="style",
        args={
            "color": Optional(str),
            "background": Optional(str),
        }
    )

    @on_render()
    async def render(self, args):
        mds = []

        if (clr := args.color) != None:
            mds.append(colors[clr])
        if (bkg := args.background) != None:
            mds.append(backgrounds[bkg])

        self.md = md(*mds)

        # allows for nesting sytles without losing 
        # effects by a nested END terminator
        cont = None
        if (style := self.first_ancestor('style')):
            cont = style.md

        return self.md + await self.render_content() + (cont or END)
            


import discord
import asyncio

from discord.ext import commands
from types import CoroutineType

from .hooks import HookTrigger
from .hooks import ReactionHook, MessageHook

    
class DiscordListener(commands.Cog):

    active_renderers = []
    bot = None

    def __init__(self, bot=None):
        if bot:
            type(self).bot = bot

    def add_renderer(cls, renderer_to_add):
        for renderer in cls.active_renderers:
            if renderer == renderer_to_add:
                return False

        # new renderers get added to the front of the list,
        # this way they have priority when receiving input
        cls.active_renderers.insert(0, renderer_to_add)
        return True

    def remove_renderer(cls, renderer_to_remove):
        for index, renderer in enumerate(cls.active_renderers):
            if renderer == renderer_to_remove:

                cls.active_renderers.pop(index)
                return True

        return False

    def get_renderers(cls, func) -> list:
        renderers = []

        for renderer in cls.active_renderers:
            if func(renderer):
                renderers.append(renderer)

        return renderers

    async def on_event(cls, trigger, renderer_filter, cleanup):
        resolved_trigger = None

        if trigger.user == cls.bot.user:
            return

        for renderer in cls.get_renderers(renderer_filter):

            # hook matching is now independent of event execution,
            # so we can use it synchronously to check all renderers before resolving the trigger
            if renderer.match_trigger(trigger):
                resolved_trigger = await renderer.resolve_trigger(trigger)
                # comment out if you want a single trigger to affect multiple templates
                break

        # this attempts to delete the reactions even when the template is discarded, 
        # leading to 404 from discord, executing events after cleanup would solve this;
        # might be good now actually
        if resolved_trigger and resolved_trigger.delete_input():
            await cleanup()
        else:
            trigger.resolve()

    def reaction_cleanup(cls, reaction, user):
        async def wrapper():
            await reaction.remove(user)

        return wrapper

    @commands.Cog.listener('on_reaction_add')
    async def on_reaction(cls, reaction: discord.Reaction, user: discord.User):

        trigger = HookTrigger(ReactionHook, data=str(reaction.emoji), user=user)

        await cls.on_event(
            trigger,
            lambda r: reaction.message.id in r.messages, 
            cls.reaction_cleanup(reaction, user)
        )

    def message_cleanup(cls, message):
        async def wrapper():
            await message.delete()

        return wrapper 

    @commands.Cog.listener('on_message')
    async def on_message(cls, message: discord.Message):

        trigger = HookTrigger(MessageHook, data=message.content, user=message.author)

        await cls.on_event(
            trigger,
            lambda r: message.channel.id == r.channel.id, 
            cls.message_cleanup(message)
        )




def similarity(sample1, sample2):
    connections = []
    for character1 in sample1:
        for i, character2 in enumerate(sample2[::-1]):
            index = len(sample2) - i - 1
            if character1 == character2:
                # finds the highest position within connection layers where index can be inserted
                if connections:
                    # goes through all the connection layers in a reverse order
                    for depth, layer in enumerate(connections[::-1]):
                        if min(layer) < index:
                            layer_index = len(connections) - depth
                            if layer_index < len(connections):
                                if index not in connections[layer_index]:
                                    connections[layer_index].append(index)
                            else:
                                connections.append([index])
                            break
                    else: 
                        if index not in connections[0]:
                            connections[0].append(index)
                else:
                    # if connection list is empty create a new connection and insert first (lowest) found value
                    connections.append([index])
            # print(character1, character2)
            # print(connections)
        
    # print(connections)

    return (len(connections) / len(sample1)) / (len(sample1) + len(sample2) - 2 * len(connections) + 1)


def find_item_in_array(string, array, threshold=0):
    value = 0
    out = None
    for item in array:
        sim = similarity(string.lower(), item.lower())
        if (value == 0 or sim > value) and sim >= threshold:
            value = sim
            out = item
    return out


if __name__ == '__main__':
    s1 = "Fliqqrrrr"
    s2 = "Flqqrrrri"
    print(similarity(s1, s2))